from syned.beamline.shape import SurfaceShape
from syned.beamline.optical_element_with_surface_shape import OpticalElementsWithSurfaceShape

class DiffractionGeometry:
    BRAGG = 0
    LAUE = 1

class Crystal(OpticalElementsWithSurfaceShape):
    """
    Constructor.

    Parameters
    ----------
    name : str, optional
        The name of the optical element.
    surface_shape : instance of SurfaceShape, optional
        The geometry of the crystal surface. if None, it is initialized to SurfaceShape().
    boundary_shape : instance of BoundaryShape, optional
        The geometry of the slit aperture. if None, it is initialized to BoundaryShape().
    material : str, optional
        The material name.
    diffraction_geometry : int (as defined in DiffractionGeometry, optional
        BRAGG = 0, LAUE = 1.
    miller_index_h : int, optional
        The Miller index H.
    miller_index_k : int, optional
        The Miller index K.
    miller_index_l : int, optional
        The Miller index L.
    asymmetry_angle : float, optional
        The asymmetry angle in rad.
    thickness : float, optional
        The crystal thickness in m.

    """
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(), # TODO: this should be None
                 boundary_shape=None,
                 material="Si",
                 diffraction_geometry=DiffractionGeometry.BRAGG,
                 miller_index_h=1,
                 miller_index_k=1,
                 miller_index_l=1,
                 asymmetry_angle=0.0,
                 thickness=0.0,
                 ):
        super().__init__(name, surface_shape, boundary_shape)
        self._material = material
        self._diffraction_geometry = diffraction_geometry
        self._miller_index_h = miller_index_h
        self._miller_index_k = miller_index_k
        self._miller_index_l = miller_index_l
        self._asymmetry_angle = asymmetry_angle
        self._thickness = thickness

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                "Name" ,                  "" ),
                    ("surface_shape",       "Surface Shape" ,         "" ),
                    ("boundary_shape",      "Boundary Shape" ,        "" ),
                    ("material",            "Material (name)" ,       "" ),
                    ("diffraction_geometry","Diffraction Geometry",   "" ),
                    ("miller_index_h",      "Miller index h",         "" ),
                    ("miller_index_k",      "Miller index k",         "" ),
                    ("miller_index_l",      "Miller index l",         "" ),
                    ("asymmetry_angle",     "Asymmetry angle",        "rad"),
                    ("thickness",           "Thickness",              "m"),
            ] )





from syned.beamline.beamline import Beamline

class WidgetDecorator(object):
    """
    Definition of a widget decorator (to be used by widget implementations).

    """

    @classmethod
    def syned_input_data(cls):
        """
        A string to help defining SYNED data in OASYS.

        Returns
        -------
        list
            [("SynedData", Beamline, "receive_syned_data")]

        """
        return [("SynedData", Beamline, "receive_syned_data")]

    @classmethod
    def append_syned_input_data(cls, inputs):
        """

        Parameters
        ----------
        inputs


        """
        for input in WidgetDecorator.syned_input_data():
            inputs.append(input)

    def receive_syned_data(self, data):
        """
        To be implemented in the main object.

        Parameters
        ----------
        data

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError("Should be implemented")



from syned.util.json_tools import load_from_json_file

from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.beamline.optical_elements.ideal_elements.screen import Screen
from syned.beamline.optical_elements.absorbers.slit import Slit

from syned.storage_ring.light_source import LightSource

from syned.beamline.beamline import Beamline
from syned.beamline.beamline_element import BeamlineElement
from syned.beamline.element_coordinates import ElementCoordinates

from syned.beamline.shape import MultiplePatch

# this is an example of a double slit (a slit with two rectangular apertures).
if __name__ == "__main__":


    # source
    src1 = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=6.0,current=0.2)
    src2 = Undulator()

    lightsource1 = LightSource("test_source",src1, src2)

    # check file o/i for test
    lightsource1.to_json("tmp.json")
    tmp = load_from_json_file("tmp.json")
    print("returned class: ",type(tmp))
    print(lightsource1.to_dictionary())
    print(tmp.to_dictionary())
    assert (lightsource1.to_dictionary() == tmp.to_dictionary())

    #optical elements
    patches = MultiplePatch()
    patches.append_rectangle(-0.02,-0.01,-0.001,0.001)
    patches.append_rectangle(0.01,0.02,-0.001,0.001)
    slit1 = Slit(name="slit1", boundary_shape=patches)

    #  check file o/i for test individual elements
    mylist = [src1, src2, slit1]
    for i,element in enumerate(mylist):
        element.to_json("tmp_%d.json"%i)

    for i, element in enumerate(mylist):
        print("\nloading element %d"%i)
        tmp = load_from_json_file("tmp_%d.json"%i)
        print("returned class: ",type(tmp))
        print(mylist[i].to_dictionary())
        print(tmp.to_dictionary())
        assert (mylist[i].to_dictionary() == tmp.to_dictionary())

    # test Beamline

    bl_slit1 = BeamlineElement(optical_element=slit1, coordinates=ElementCoordinates(p=10.0,q=3.0))
    BL = Beamline(light_source=lightsource1, beamline_elements_list=[bl_slit1])

    # check file o/i for test full beamline
    BL.to_json("tmp_bl.json")
    tmp = load_from_json_file("tmp_bl.json")
    print("returned class: ",type(tmp))
    print(BL.to_dictionary())
    print(tmp.to_dictionary())
    assert(BL.to_dictionary() == tmp.to_dictionary())

    print(BL.info())



################## LSBL Reports #################################################################

# Yashchuk, V. V., “An analytical solution for shape of diaboloid mirror,” Light
# Source Beam Line Note LSBL-1436, Advanced Light Source, Berkeley (January 8, 2020).

# Yashchuk, V. V., “Shape of diaboloid mirror in laboratory coordinate system,”
# Light Source Beam Line Note LSBL-1437c, Advanced Light Source, Berkeley (January 13, 2020).

# K. Goldberg and Manuel Sanchez del Rio, “Direct Solution of
# Diaboloid Mirrors,” Light Source Beam Line Note LSBL- 1440 (ALS, Berkeley, February 18, 2020).

# Lacey, I., Sanchez del Rio, M., and Yashchuk, V. V., “Analytical expression for
# the diaboloid shape in laboratory mirror coordinates verified by ray-tracing simulations,” Light
# Source Beam Line Note LSBL-1445, Advanced Light Source, Berkeley (March 24, 2020).

# Yashchuk, V. V., “Diaboloid shape approximation with a sagittal conical cylinder
# bent to a tangential parabola: Analytical consideration,” Light Source Beam Line Note LSBL-1451,
# Advanced Light Source, Berkeley (April 14, 2020).

# Yashchuk, V. V., “Explicit algebraic derivation of an expression for the exact
# shape of diaboloid mirror in laboratory coordinate system,”
# Light Source Beam Line Note LSBL˗1462, Advanced Light Source, Berkeley (May 17, 2020).

# H. A. Padmore, “Sagittal shape difference between a cylinder and a diaboloid at 2:1
# demagnification,” Light Source Beam Line Note LSBL-1465, Advanced Light Source, Berkeley
# (June 03, 2020).

################## Papers #################################################################

# W. R. McKinney, J. M. Glossinger, H. A. Padmore, and M. R. Howells,
# "Optical path function calculation for an incoming cylindrical wave," Proc. SPIE. 7448, 744809/1-8
# (2009); https://doi.org/10.1117/12.828490.

# V. V. Yashchuk, I. Lacey, and M. Sanchez del Rio, “Analytical expressions
# of the surface shape of ‘diaboloid’ mirrors,” Proc. SPIE 11493, 114930N/1-13 (2020);
# https://doi.org/10.1117/12.2568332.



import numpy

from syned.tools.diaboloid.fqs import single_quartic
from syned.tools.diaboloid.fqs import quartic_roots

def diaboloid_approximated_point_to_segment(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001),
        detrend=0):
    """
    Create a numeric mesh of the mirror using the approximated equation of diaboloid (point to segment)
    References:

    Kenneth Goldberg & M. Sanchez del Rio
    "Direct Solution of Diaboloid Mirrors"
    LSBL-1440 (February 2020)

    M. Sanchez del Rio, K. Goldberg, V. Yashchuk, I. Lacey and H. Padmore
    Simulations of applications using diaboloid mirrors
    Journal of Synchrotron Radiation (submitted 2021)

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """
    X = numpy.outer(x, numpy.ones_like(y))
    Y = numpy.outer(numpy.ones_like(x), y)

    s = p * numpy.cos(2 * theta)
    z0 = p * numpy.sin(2 * theta)
    c = p + q

    # this is Eq. 12 in LSBL-1440 (a bit reorganised)
    Z = - numpy.sqrt(c ** 2 + q ** 2 - s ** 2 - 2 * Y * (s + q) - 2 * c * numpy.sqrt(X ** 2 + (q - Y) ** 2))
    Z += z0

    if detrend == 0:
        zfit = 0
    elif detrend == 1:
        zfit = -theta * y
    elif detrend == 2:
        zcentral = Z[Z.shape[0] // 2, :]
        zcoeff = numpy.polyfit(y[(y.size // 2 - 10):(y.size // 2 + 10)],
                               zcentral[(y.size // 2 - 10):(y.size // 2 + 10)], 1)
        zfit = zcoeff[1] + y * zcoeff[0]

    for i in range(Z.shape[0]):
        Z[i, :] = Z[i, :] - zfit

    return Z, X, Y


def diaboloid_approximated_segment_to_point(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001),
        detrend=0):
    """
    Create a numeric mesh of the mirror using the approximated equation of diaboloid (segmet to point)
    References:

    Kenneth Goldberg & M. Sanchez del Rio
    "Direct Solution of Diaboloid Mirrors"
    LSBL-1440 (February 2020)

    M. Sanchez del Rio, K. Goldberg, V. Yashchuk, I. Lacey and H. Padmore
    Simulations of applications using diaboloid mirrors
    Journal of Synchrotron Radiation (submitted 2021)

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """

    Z, X, Y = diaboloid_approximated_point_to_segment(p=q, q=p, theta=theta, x=x, y=y,
                                              detrend=detrend)
    for i in range(x.size):
        Z[i,:] = numpy.flip(Z[i,:])

    return Z, X, Y

def toroid_point_to_segment(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001)):
    """

    Create a numeric mesh of the toroid (point to segment)

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """

    Rt = 2.0 / numpy.sin(theta) / (1 / p)

    Rs = 2.0 * numpy.sin(theta) / (1 / p + 1 / q)

    print("Toroid Rt: %9.6f m, Rs: %9.6f m" % (Rt, Rs))

    height_tangential = Rt - numpy.sqrt(Rt ** 2 - y ** 2)
    height_sagittal = Rs - numpy.sqrt(Rs ** 2 - x ** 2)

    Z = numpy.zeros((x.size, y.size))

    for i in range(x.size):
        Z[i,:] = height_tangential

    for i in range(y.size):
        Z[:,i] += height_sagittal

    return Z

def toroid_segment_to_point(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001)):
    """

    Create a numeric mesh of the toroid (segment to point)

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """

    Z = toroid_point_to_segment(p=q, q=p, theta=theta, x=x, y=y)
    for i in range(x.size):
        Z[i,:] = numpy.flip(Z[i,:])
    return Z

def parabolic_cone_point_to_segment(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001)):
    """
    Create a numerical mesh for an approximated diaboloid (point to segment focusing)
    as calculated by Valeriy Yashchuk

    Valeriy V. Yashchuk
    Diaboloid shape approximation with a sagittal conical cylinder bent to a tangential parabola:
    Analytical consideration
    Light Source Beam Line Note LSBL-1451, Advanced Light Source, Berkeley (April 14, 2020).


    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """
    X = numpy.outer(x, numpy.ones_like(y))
    Y = numpy.outer(numpy.ones_like(x), y)

    c = numpy.cos(theta)
    s = numpy.sin(theta)
    c2 = numpy.cos(2 * theta)
    s2 = numpy.sin(2 * theta)
    pq = p + q

    # Equation 15 in V. Yashchuk LSBL 1451
    k1 = p * q * c * s2 / pq
    k2 = s2 * (q - 2 * p * c**2 ) / 2 / pq
    Z = Y * s / c - \
        2  * s / c**2 * numpy.sqrt(Y * p * c + p**2) + \
        2 * p * s / c**2 + \
        k1 + k2 * Y \
        - numpy.sqrt( (k1 + k2 * Y)**2 - X**2 )

    return Z, X, Y

def parabolic_cone_segment_to_point(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001)):
    """
    Create a numerical mesh for an approximated diaboloid (point to segment focusing)
    as calculated by Valeriy Yashchuk

    Valeriy V. Yashchuk
    Diaboloid shape approximation with a sagittal conical cylinder bent to a tangential parabola:
    Analytical consideration
    Light Source Beam Line Note LSBL-1451, Advanced Light Source, Berkeley (April 14, 2020).

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """
    Z, X, Y = parabolic_cone_point_to_segment(p=q, q=p, theta=theta, x=x, y=y)
    for i in range(x.size):
        Z[i,:] = numpy.flip(Z[i,:])

    return Z, X, Y


def parabolic_cone_linearized_point_to_segment(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001)):
    """

    Create a numerical mesh for the linearized approximated diaboloid (point to segment focusing)
    as calculated by Valeriy Yashchuk

    References:

    Valeriy V. Yashchuk
    Diaboloid shape approximation with a sagittal conical cylinder bent to a tangential parabola:
    Analytical consideration
    Light Source Beam Line Note LSBL-1451, Advanced Light Source, Berkeley (April 14, 2020).

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """
    X = numpy.outer(x, numpy.ones_like(y))
    Y = numpy.outer(numpy.ones_like(x), y)

    c = numpy.cos(theta)
    s = numpy.sin(theta)
    c2 = numpy.cos(2 * theta)
    s2 = numpy.sin(2 * theta)
    pq = p + q

    # use the central meridional profile like in parabolic_cone_point_to_segment()
    Z = Y * s / c - 2  * s / c**2 * numpy.sqrt(Y * p * c + p**2) + 2 * p * s / c**2 \
        - numpy.sqrt( \
        (p * q * c * s2 / pq + s2 * (q - 2 * p * c**2 ) / 2 / pq * Y)**2 - (X*0)**2) + \
        p * q * c * s2 / pq + s2 * (q - 2 * p * c**2) / 2 / pq * Y

    # we add now the sagittal profile with radius calculated using Eq 11 in LSBL 1451
    for j in range(y.size):
        # Rs = p * q * numpy.sin(2 * theta) / (p + q) # TODO missing c:  = 2 p * q * sin(theta) * cos^2(theta)
        # Rs += (q * numpy.tan(theta) - 2 * p * numpy.sin(theta) * numpy.cos(theta)) / (p + q) * y[j]
        Rs = p * q * c * numpy.sin(2 * theta) / (p + q)
        Rs += numpy.sin(2 * theta) * (q  - 2 * p * c**2) / 2 / (p + q) * y[j]
        height_sagittal = Rs - numpy.sqrt(Rs ** 2 - x ** 2)
        print("y=%f Rs=%f" % (y[j], Rs))
        Z[:,j] += height_sagittal

    return Z, X, Y

def parabolic_cone_linearized_segment_to_point(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001)):
    """

    Create a numerical mesh for the linearized approximated diaboloid (segment to point focusing)
    as calculated by Valeriy Yashchuk

    Valeriy V. Yashchuk
    Diaboloid shape approximation with a sagittal conical cylinder bent to a tangential parabola:
    Analytical consideration
    Light Source Beam Line Note LSBL-1451, Advanced Light Source, Berkeley (April 14, 2020).

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """
    Z, X, Y = parabolic_cone_linearized_point_to_segment(p=q, q=p, theta=theta, x=x, y=y)
    for i in range(x.size):
        Z[i,:] = numpy.flip(Z[i,:])

    return Z, X, Y

def diaboloid_exact_point_to_segment(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001),
        ):
    """

    Create a numerical mesh for the exact diaboloid (point to segment focusing) solving
    a fourth degree equation, as calculated by Valeriy Yashchuk

    References:

    Valeriy V. Yashchuk
    Explicit algebraic derivation of an expression for the exact shape of diaboloidal
    mirror in laboratory coordinate system
    Light Source Beam Line Note LSBL˗1462, Advanced Light Source, Berkeley (May 17, 2020).

    Yashchuk, V. V., Goldberg, K., Lacey, I., McKinney, W. R., Sanchez del Rio, M. & Padmore,H.
    Journal of Synchrotron Radiation, submitted(2021).

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """

    X = numpy.outer(x, numpy.ones_like(y))
    Y = numpy.outer(numpy.ones_like(x), y)

    c = numpy.cos(theta)
    s = numpy.sin(theta)

    c2 = numpy.cos(2 * theta)
    s2 = numpy.sin(2 * theta)

    # Coefficients after Valeriy V. Yashchuk LSBL-1462 May 2020
    # A = −Cos[θ] 4 .
    # B = 4(r1 − r2)Cos[θ] 2 Sin[θ] + 4Cos[θ] 3 Sin[θ]z;
    # C = 4r2((r1 + r2)Cos[θ] 2 + 4r1Sin[θ] 2 ) + 2Cos[θ](−3r1 + r2 + (r1 − 3r2)Cos[2θ])z − 6(Cos[θ] 2 Sin[θ] 2 )z 2 ;
    # D = −16r1r2(r1 + r2)Sin[θ] + 4(r1 + r2)(2r1 − r2)Sin[2θ]z + 2(3r1 + r2 + (r1 + 3r2)Cos[2θ])Sin[θ]z 2 + 4Cos[θ]Sin[θ] 3 z 3 ;
    # E = 4(r1 + r2) 2 x 2 + 4r2(r1 + r2)Sin[θ] 2 z 2 − 4((r1 + r2)Cos[θ]Sin[θ] 2 )z 3 − Sin[θ] 4 z 4 ;

    A = -c**4 * numpy.ones_like(X)
    B = 4 * (p - q) * c**2 * s \
                + 4 * c**3 * s * Y
    C = 4 * q * ( (p + q) * c**2 + 4 * p * s**2 ) \
                + 2 * c * (q - 3 * p + (p - 3 * q) * c2) * Y \
                - 6 * c**2 * s**2 * Y**2
    D = -16 * p * q * (p + q) * s \
                + 4 * (p + q) * (2 * p - q) * s2 * Y \
                + 2 * (3 * p + q + (3 * q + p) * c2) * s * Y**2 \
                + 4 * c * s**3 * Y**3
    E = 4 * (p + q)**2 * X**2 \
            + 4 * q * (p + q) * s**2 * Y**2 \
            - 4 * (p + q) * c * s**2 * Y**3 \
            - s**4 * Y**4

    # get good solution: the one that is zero at (0,0)
    ix = x.size // 2
    iy = y.size // 2
    solutions = single_quartic(A[ix, iy], B[ix, iy], C[ix, iy], D[ix, iy], E[ix, iy])
    aa = []
    for sol in solutions:
        if numpy.abs(sol.imag) < 1e-15:
            aa.append(numpy.abs(sol.real))
        else:
            aa.append(1e10)
    isel = numpy.argmin(aa)


    # calculate solutions array
    P = numpy.zeros((A.size, 5))
    P[:, 0] = A.flatten()
    P[:, 1] = B.flatten()
    P[:, 2] = C.flatten()
    P[:, 3] = D.flatten()
    P[:, 4] = E.flatten()
    SOLUTION = quartic_roots(P)

    # return result
    SOLUTION_GOOD = (SOLUTION[:,isel]).flatten()
    SOLUTION_GOOD.shape = A.shape
    Z = SOLUTION_GOOD.real
    return Z, X, Y

def diaboloid_exact_segment_to_point(
        p=29.3,
        q=19.53,
        theta=4.5e-3,
        x=numpy.linspace(-0.01, 0.01, 101),
        y=numpy.linspace(-0.1, 0.1, 1001),
        ):
    """

    Create a numerical mesh for the exact diaboloid (segment to point focusing) solving
    a fourth degree equation, as calculated by Valeriy Yashchuk

    References:

    Valeriy V. Yashchuk
    Explicit algebraic derivation of an expression for the exact shape of diaboloidal
    mirror in laboratory coordinate system
    Light Source Beam Line Note LSBL˗1462, Advanced Light Source, Berkeley (May 17, 2020).

    Yashchuk, V. V., Goldberg, K., Lacey, I., McKinney, W. R., Sanchez del Rio, M. & Padmore,H.
    Journal of Synchrotron Radiation, submitted(2021).

    :param p: distance source to mirror [m]
    :param q: distance mirror to focus [m]
    :param theta: grazing incidence angle [rad]
    :param y: x (sagittal) array
    :param y: y (tangential) array
    :return: Z, X, Y
    """
    Z, X, Y = diaboloid_exact_point_to_segment(p=q, q=p, theta=theta, x=x, y=y)
    for i in range(x.size):
        Z[i,:] = numpy.flip(Z[i,:])

    return Z, X, Y


from syned.beamline.shape import SurfaceShape
from syned.beamline.optical_element_with_surface_shape import OpticalElementsWithSurfaceShape

class Interface(OpticalElementsWithSurfaceShape):
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(),
                 boundary_shape=None,
                 material_object=None,
                 material_image=None,):
        """
        Defines an interface (a surface with different materials in side 1 and side 2).

        Parameters
        ----------
        name : str, optional
            The name of the optical element.
        surface_shape : instance of SurfaceShape, optional
            The geometry of the crystal surface. if None, it is initialized to SurfaceShape().
        boundary_shape : instance of BoundaryShape, optional
            The geometry of the slit aperture. if None, it is initialized to BoundaryShape().
        material_object : str, optional
            The material in side 1 (object side).
        material_image : str, optional
            The material in side 2 (object image).

        """

        super().__init__(name, surface_shape, boundary_shape)
        self._material_object = material_object
        self._material_image = material_image
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                "Name" ,                               "" ),
                    ("surface_shape",       "Surface shape",                       "" ),
                    ("boundary_shape",      "Boundary shape",                      "" ),
                    ("material_object",     "Material in object side (element, compound, name or refraction index)", "" ),
                    ("material_image",      "Material in image side (element, compound, name or refraction index)", ""),
           ] )

    def get_material_object(self):
        """
        Returns the material of side 1 (object).

        Returns
        -------
        str

        """
        return self._material_object

    def get_material_image(self):
        """
        Returns the material of side 2 (image).

        Returns
        -------
        str

        """
        return self._material_image


from syned.beamline.optical_element_with_surface_shape import OpticalElementsWithMultipleShapes
from syned.beamline.shape import Plane

class Lens(OpticalElementsWithMultipleShapes):
    def __init__(self,
                 name="Undefined",
                 surface_shape1=None,
                 surface_shape2=None,
                 boundary_shape=None,
                 material="",
                 thickness=0.0):
        """
        Defines a lens. It is composed by two surfaces (surface_shape1 and surface_shape2) and a material
        within them.

        Parameters
        ----------
        name : str, optional
            The name of the optical element.
        surface_shape1 : instance of SurfaceShape, optional
            The geometry of the lens surface 1. if None, it is initialized to SurfaceShape().
        surface_shape2 : instance of SurfaceShape, optional
            The geometry of the lens surface 2. if None, it is initialized to SurfaceShape().
        boundary_shape : instance of BoundaryShape, optional
            The geometry of the slit aperture. if None, it is initialized to BoundaryShape().
        material : str
            A string defining the material within the two surfaces.
        thickness : float
            The distance between the two surfaces at the center of the lens in m.

        """

        if surface_shape1 is None: surface_shape1 = Plane()
        if surface_shape2 is None: surface_shape2 = Plane()

        super(Lens, self).__init__(name=name,
                                   boundary_shape=boundary_shape,
                                   surface_shapes=[surface_shape1, surface_shape2])
        self._material = material
        self._thickness = thickness

        # support text contaning name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                "Name" ,                                "" ),
                    ("surface_shape1",      "Surface shape 1",                      "" ),
                    ("surface_shape2",      "Surface shape 2",                      ""),
                    ("boundary_shape",      "Boundary shape",                       "" ),
                    ("material",            "Material (element, compound or name)", "" ),
                    ("thickness",           "Thickness",                            "m"),
            ] )

    def get_thickness(self):
        """
        Returns the lens thickness in m.

        Returns
        -------
        float

        """
        return self._thickness

    def get_material(self):
        """
        Returns the lens material.

        Returns
        -------
        str

        """
        return self._material

    def get_boundary_shape(self):
        """
        Returns the boundary shape.

        Returns
        -------
        instance of BoundaryShape

        """
        return self._boundary_shape

    def get_surface_shape1(self):
        """
        Returns the shape of surface 1.

        Returns
        -------
        instance of SurfaceShape

        """
        return self.get_surface_shape(index=0)

    def get_surface_shape2(self):
        """
        Returns the shape of surface 2.

        Returns
        -------
        instance of SurfaceShape

        """
        return self.get_surface_shape(index=1)




from syned.storage_ring.magnetic_structures.insertion_device import InsertionDevice

class Wiggler(InsertionDevice):
    """
    Constructor.

    Parameters
    ----------
    K_vertical : float, optional
        The deflection K parameter corresponding to magnetic field in the vertical direction.
    K_horizontal : float, optional
        The deflection K parameter corresponding to magnetic field in the horizontal direction.
    period_length : float, optional
        The ID period in m.
    number_of_periods : float, optional
        The number of periods. It may be a float, considering that number_of_periods = ID_length / period_length.

    """

    def __init__(self, K_vertical = 0.0, K_horizontal = 0.0,period_length = 0.0, number_of_periods = 1):
        InsertionDevice.__init__(self,
                                 K_vertical=K_vertical,
                                 K_horizontal=K_horizontal,
                                 period_length=period_length,
                                 number_of_periods=number_of_periods)





from syned.beamline.optical_elements.ideal_elements.ideal_element import IdealElement

class IdealLens(IdealElement):
    """
    Defines an ideal lens. It converts a plane wave into:
    * an spherical converging wave (if focal_x=focal_y).
    * a toroidal converging wave (if focal_x != focal_y).
    * a cylindrical wave (if focal_x or focal_y is infinity).

    Constructor.

    Parameters
    ----------
    name : str, optional
        The name of the optical element.
    focal_x : float
        The focal length in meters along the X direction.
    focal_y : float
        The focal length in meters along the Y direction.

    """
    def __init__(self, name="Undefined", focal_x=1.0, focal_y=1.0):
        IdealElement.__init__(self, name=name)
        self._focal_x = focal_x
        self._focal_y = focal_y
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name"          , "Name"                          , ""),
                    ("boundary_shape", ""                              , ""),
                    ("focal_x"       , "Focal length in x [horizontal]", "m" ),
                    ("focal_y"       , "Focal length in y [vertical]"  , "m" ),
            ] )

    def focal_x(self):
        """
        Returns the focal length in the X direction.

        Returns
        -------
        float

        """
        return self._focal_x

    def focal_y(self):
        """
        Returns the focal length in the Y direction.

        Returns
        -------
        float

        """
        return self._focal_y



"""
Base class for electron beams.

This class is intentionally shorten for simplicity.
Usually we would need to consider also the electron distribution within the beam.
"""

from syned.syned_object import SynedObject
import scipy.constants as codata
import numpy

class ElectronBeam(SynedObject):
    def __init__(self,
                 energy_in_GeV = 1.0,
                 energy_spread = 0.0,
                 current = 0.1,
                 number_of_bunches = 400, # TODO: not used, to be removed (with care!)
                 moment_xx=0.0,
                 moment_xxp=0.0,
                 moment_xpxp=0.0,
                 moment_yy=0.0,
                 moment_yyp=0.0,
                 moment_ypyp=0.0):
        """
        Defines an electron beam at a given point of the storage ring.

        Parameters
        ----------
        energy_in_GeV : float, optional
            The electron energy in GeV.
        energy_spread : float, optional
            The electron energy spread (in a fraction of the energy_in_GeV).
        current : float, optional
            The electron beam current intensity in A.
        number_of_bunches : float, optional
            The number of bunches in the storage ring.
        moment_xx : float, optional
            The <x^2> moment.
        moment_xxp : float, optional
            The <x x'> moment.
        moment_xpxp : float, optional
            The <x'^2> moment.
        moment_yy : float, optional
            The <y^2> moment.
        moment_yyp : float, optional
            The <y y'> moment.
        moment_ypyp : float, optional
            The <y'^2> moment.
        """


        self._energy_in_GeV       = energy_in_GeV
        self._energy_spread       = energy_spread
        self._current             = current
        self._number_of_bunches   = number_of_bunches

        self._moment_xx           = moment_xx
        self._moment_xxp          = moment_xxp
        self._moment_xpxp         = moment_xpxp
        self._moment_yy           = moment_yy
        self._moment_yyp          = moment_yyp
        self._moment_ypyp         = moment_ypyp

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("energy_in_GeV"      , "Electron beam energy"                  , "GeV" ),
                    ("energy_spread"      , "Electron beam energy spread (relative)", ""    ),
                    ("current"            , "Electron beam current"                 , "A"   ),
                    ("number_of_bunches"  , "Number of bunches"                     , ""    ),
                    ("moment_xx"          , "Moment (spatial^2, horizontal)"        , "m^2" ),
                    ("moment_xxp"         , "Moment (spatial-angular, horizontal)"  , "m"   ),
                    ("moment_xpxp"        , "Moment (angular^2, horizontal)"        , ""    ),
                    ("moment_yy"          , "Moment (spatial^2, vertical)"          , "m^2" ),
                    ("moment_yyp"         , "Moment (spatial-angular, vertical)"    , "m"   ),
                    ("moment_ypyp"        , "Moment (angular^2, vertical)"          , ""    ),
            ] )



    #
    # initializares
    #
    @classmethod
    def initialize_as_pencil_beam(cls, energy_in_GeV = 1.0, energy_spread = 0.0, current = 0.1, **params):
        """
        Creates an electron pencil beam.

        Parameters
        ----------
        energy_in_GeV : float, optional
            The electron energy in GeV.
        energy_spread : float, optional
            The electron energy spread (in a fraction of the energy_in_GeV).
        current : float, optional
            The electron beam current intensity in A.
        params :
            other keyword parameters accepted by ElectronBeam (if used, the result may not be a pencil beam.)

        Returns
        -------
        instance of ElectronBeam

        """
        return cls(energy_in_GeV=energy_in_GeV,
                   energy_spread=energy_spread,
                   current=current,
                   number_of_bunches=1,
                   **params)


    #
    # useful getters
    #
    def get_sigmas_real_space(self):
        """
        Returns the sigmas in real space.

        Returns
        -------
        tuple
            (sigma_x, sigma_y)

        """
        return numpy.sqrt(self._moment_xx),\
               numpy.sqrt(self._moment_yy)

    def get_sigmas_divergence_space(self):
        """
        Returns the sigmas in divergence (angle) space.

        Returns
        -------
        tuple
            (sigma_x', sigma_y')

        """
        return numpy.sqrt(self._moment_xpxp),\
               numpy.sqrt(self._moment_ypyp)

    def get_sigmas_horizontal(self):
        """
        Returns the sigmas in horizontal direction.

        Returns
        -------
        tuple
            (sigma_x, sigma_x')

        """
        return numpy.sqrt(self._moment_xx),\
               numpy.sqrt(self._moment_xpxp)

    def get_sigmas_vertical(self):
        """
        Returns the sigmas in vertical direction.

        Returns
        -------
        tuple
            (sigma_y, sigma_y')

        """
        return numpy.sqrt(self._moment_yy),\
               numpy.sqrt(self._moment_ypyp)

    def get_sigmas_all(self):
        """
        Returns all sigmas.

        Returns
        -------
        tuple
            (sigma_x, sigma_x', sigma_y, sigma_y')

        """
        return numpy.sqrt(self._moment_xx),\
               numpy.sqrt(self._moment_xpxp),\
               numpy.sqrt(self._moment_yy),\
               numpy.sqrt(self._moment_ypyp)

    def get_moments_horizontal(self):
        """
        Returns the moments in the horizontal direction.

        Returns
        -------
        tuple
            ( <x^2>, <x x'>, <x'^2>)

        """
        return self._moment_xx, self._moment_xxp, self._moment_xpxp

    def get_moments_vertical(self):
        """
        Returns the moments in the vertical direction.

        Returns
        -------
        tuple
            ( <y^2>, <y y'>, <y'^2>)

        """
        return self._moment_yy, self._moment_yyp, self._moment_ypyp

    def get_moments_all(self):
        """
        Returns all moments.

        Returns
        -------
        tuple
            ( <x^2>, <x x'>, <x'^2>, <y^2>, <y y'>, <y'^2>)

        """
        return self._moment_xx, self._moment_xxp, self._moment_xpxp, self._moment_yy, self._moment_yyp, self._moment_ypyp

    def get_twiss_no_dispersion_horizontal(self):
        """
        Returns the Twiss parameters in horizontal direction.
        (The energy disperion is not considered.)

        Returns
        -------
        tuple
            (emittance_x, alpha_x, beta_x).

        """
        emittance_x = numpy.sqrt(self._moment_xx * self._moment_xpxp - self._moment_xxp**2)
        alpha_x = -self._moment_xxp / emittance_x
        beta_x = self._moment_xx / emittance_x
        return emittance_x, alpha_x, beta_x

    def get_twiss_no_dispersion_vertical(self):
        """
        Returns the Twiss parameters in vertical direction.
        (The energy disperion is not considered.)

        Returns
        -------
        tuple
            (emittance_y, alpha_y, beta_y).

        """
        emittance_y = numpy.sqrt(self._moment_yy * self._moment_ypyp - self._moment_yyp**2)
        alpha_y = -self._moment_yyp / emittance_y
        beta_y = self._moment_yy / emittance_y
        return emittance_y,alpha_y,beta_y

    def get_twiss_no_dispersion_all(self):
        """
        Returns all Twiss parameters.
        (The energy disperion is not considered.)

        Returns
        -------
        tuple
            (emittance_x, alpha_x, beta_x, emittance_y, alpha_y, beta_y).

        """
        ex, ax, bx =  self.get_twiss_no_dispersion_horizontal()
        ey, ay, by = self.get_twiss_no_dispersion_vertical()
        return ex, ax, bx, ey, ay, by

    def energy(self):
        """
        Returns the electron energy in GeV.

        Returns
        -------
        float

        """
        return self._energy_in_GeV

    def current(self):
        """
        Returns the electron current in A.

        Returns
        -------
        float

        """
        return self._current

    #
    # setters
    #
    def set_sigmas_real_space(self, sigma_x=0.0, sigma_y=0.0):
        """
        Sets the electron beam parameters in real space from the sigma values.

        Parameters
        ----------
        sigma_x : float, optional
            The sigma in horizontal direction.
        sigma_y : float, optional
            The sigma in vertical direction.

        """
        self._moment_xx = sigma_x**2
        self._moment_yy = sigma_y**2

    def set_sigmas_divergence_space(self, sigma_xp=0.0, sigma_yp=0.0):
        """
        Sets the electron beam parameters in divergence space from the sigma values.

        Parameters
        ----------
        sigma_xp : float, optional
            The sigma in horizontal direction.
        sigma_yp : float, optional
            The sigma in vertical direction.

        """
        self._moment_xpxp = sigma_xp**2
        self._moment_ypyp = sigma_yp**2

    def set_sigmas_horizontal(self, sigma_x=0.0, sigma_xp=0.0):
        """
        Sets the electron beam parameters from the sigma values in horizontal direction.

        Parameters
        ----------
        sigma_x : float, optional
            The sigma in real space.
        sigma_xp : float, optional
            The sigma in divergence space.

        """
        self._moment_xx = sigma_x**2
        self._moment_xpxp = sigma_xp**2

    def set_sigmas_vertical(self, sigma_y=0.0, sigma_yp=0.0):
        """
        Sets the electron beam parameters from the sigma values in vertical direction.

        Parameters
        ----------
        sigma_y : float, optional
            The sigma in real space.
        sigma_yp : float, optional
            The sigma in divergence space.

        """
        self._moment_yy = sigma_y**2
        self._moment_ypyp = sigma_yp**2

    def set_sigmas_all(self, sigma_x=0.0, sigma_xp=0.0, sigma_y=0.0, sigma_yp=0.0):
        """
        Sets the electron beam parameters from the sigma values in both horizontal and vertical direction.

        Parameters
        ----------
        sigma_x : float, optional
            The sigma in real space (horizontal).
        sigma_xp : float, optional
            The sigma in divergence space (horizontal).
        sigma_y : float, optional
            The sigma in real space (vertical).
        sigma_yp : float, optional
            The sigma in divergence space (vertical).

        """
        self._moment_xx = sigma_x**2
        self._moment_xpxp = sigma_xp**2
        self._moment_yy = sigma_y**2
        self._moment_ypyp = sigma_yp**2

    def set_energy_from_gamma(self, gamma):
        """
        Sets the electron energy from the gamma value (Lorentz factor).

        Parameters
        ----------
        gamma : float

        """
        self._energy_in_GeV = (gamma / 1e9) * (codata.m_e *  codata.c**2 / codata.e)

    def set_moments_horizontal(self, moment_xx, moment_xxp, moment_xpxp):
        """
        Sets the moments in the horizontal direction.

        Parameters
        ----------
        moment_xx : float
            The <x^2> moment.
        moment_xxp : float
            The <x x'> moment.
        moment_xpxp : float,
            The <x'^2> moment.

        """
        self._moment_xx   = moment_xx
        self._moment_xxp  = moment_xxp
        self._moment_xpxp = moment_xpxp

    def set_moments_vertical(self, moment_yy, moment_yyp, moment_ypyp):
        """
        Sets the moments in the vertical direction.

        Parameters
        ----------
        moment_yy : float
            The <y^2> moment.
        moment_yyp : float
            The <y y'> moment.
        moment_ypyp : float
            The <y'^2> moment.

        """
        self._moment_yy   = moment_yy
        self._moment_yyp  = moment_yyp
        self._moment_ypyp = moment_ypyp

    def set_moments_all(self, moment_xx, moment_xxp, moment_xpxp, moment_yy, moment_yyp, moment_ypyp):
        """
        Sets the moments.

        Parameters
        ----------
        moment_xx : float
            The <x^2> moment.
        moment_xxp : float
            The <x x'> moment.
        moment_xpxp : float,
            The <x'^2> moment.
        moment_yy : float
            The <y^2> moment.
        moment_yyp : float
            The <y y'> moment.
        moment_ypyp : float
            The <y'^2> moment.

        """
        self.set_moments_horizontal(moment_xx,moment_xxp,moment_xpxp)
        self.set_moments_vertical(moment_yy,moment_yyp,moment_ypyp)

    def set_twiss_horizontal(self, emittance_x, alpha_x, beta_x, eta_x=0, etap_x=0):
        """
        Sets the electron beam parameters from the Twiss values in the horizontal direction.

        Parameters
        ----------
        emittance_x : float
            The emittance value in horizontal.
        alpha_x : float
            The alpha value in horizontal.
        beta_x : float
            The beta value in horizontal.
        eta_x : float, optional
            The eta value in horizontal.
        etap_x : float, optional
            The eta' value in horizontal.

        """
        gamma_x = (1 + alpha_x**2) / beta_x
        self._moment_xx   = beta_x   * emittance_x + eta_x**2 * self._energy_spread**2
        self._moment_xxp  = -alpha_x * emittance_x + eta_x * etap_x * self._energy_spread ** 2
        self._moment_xpxp = gamma_x  * emittance_x + etap_x**2 * self._energy_spread ** 2

    def set_twiss_vertical(self, emittance_y, alpha_y, beta_y, eta_y=0, etap_y=0):
        """
        Sets the electron beam parameters from the Twiss values in the vertical direction.

        Parameters
        ----------
        emittance_y : float
            The emittance value in vertical.
        alpha_x : float
            The alpha value in vertical.
        beta_x : float
            The beta value in vertical.
        eta_x : float, optional
            The eta value.
        etap_x : float, optional
            The eta' value in vertical.

        """
        gamma_y = (1 + alpha_y**2) / beta_y
        self._moment_yy   = beta_y   * emittance_y + eta_y**2 * self._energy_spread**2
        self._moment_yyp  = -alpha_y * emittance_y + eta_y * etap_y * self._energy_spread ** 2
        self._moment_ypyp = gamma_y  * emittance_y + etap_y**2 * self._energy_spread ** 2

    def set_twiss_all(self, emittance_x, alpha_x, beta_x, eta_x, etap_x,
                           emittance_y, alpha_y, beta_y, eta_y, etap_y):
        """
        Sets the electron beam parameters from the Twiss values.

        Parameters
        ----------
        emittance_x : float
            The emittance value in horizontal.
        alpha_x : float
            The alpha value in horizontal.
        beta_x : float
            The beta value in horizontal.
        eta_x : float
            The eta value in horizontal.
        etap_x : float
            The eta' value in horizontal.
        emittance_y : float
            The emittance value in vertical.
        alpha_x : float
            The alpha value in vertical.
        beta_x : float
            The beta value in vertical.
        eta_x : float
            The eta value.
        etap_x : float
            The eta' value in vertical.

        """
        self.set_twiss_horizontal(emittance_x,alpha_x,beta_x,eta_x,etap_x)
        self.set_twiss_vertical(emittance_y,alpha_y,beta_y,eta_y,etap_y)

    #
    # some easy calculations
    #
    def gamma(self):
        """
        returns the Gamma or Lorentz factor.

        Returns
        -------
        float

        """
        return self.lorentz_factor()

    def lorentz_factor(self):
        """
        returns the Gamma or Lorentz factor.

        Returns
        -------
        float

        """
        return 1e9 * self._energy_in_GeV / (codata.m_e *  codata.c**2 / codata.e)

    def electron_speed(self):
        """
        Returns the electron velocity in c units.

        Returns
        -------
        float

        """
        return numpy.sqrt(1.0 - 1.0 / self.lorentz_factor() ** 2)



    #
    #dictionnary interface, info etc
    #

if __name__ == "__main__":

    a = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=2.0,current=0.5, energy_spread=0.00095)

    a.set_twiss_horizontal(70e-12, 0.827, 0.34, 0, 0 ) #0.0031, -0.06)
    a.set_twiss_vertical(  70e-12, -10.7, 24.26, 0, 0.0)

    print("twiss data: 70e-12, 0.827, 0.34, 70e-12, -10.7, 24.26,")
    print("sigmas: ",a.get_sigmas_all())
    print("moments: ",a.get_moments_all())
    print("twiss: ",a.get_twiss_no_dispersion_all())


    s = a.get_sigmas_all()
    a.set_sigmas_all(s[0],s[1],s[2],s[3])

    print("\n\nsigmas: ",a.get_sigmas_all())
    print("moments: ",a.get_moments_all())
    print("twiss: ",a.get_twiss_no_dispersion_all())


#    a.to_dictionary()


    # fd = a.to_full_dictionary()
    # dict = a.to_dictionary()
    #
    # print(dict)
    #
    # for key in fd:
    #     print(key,fd[key][0])
    #
    # for key in fd:
    #     print(key,dict[key])
    #
    # print(a.keys())
    # print(a.info())



"""
Base class for an optical element.
"""
from syned.syned_object import SynedObject

class OpticalElement(SynedObject):
    """
    Constructor.

    Parameters
    ----------
    name : str
        The element name.
    boundary_shape : instance of BoundaryShape, optional
        The element shape. The default=None means no shape associated to the optical element.
    """
    def __init__(self, name="Undefined", boundary_shape=None):
        self._name = name
        self._boundary_shape = boundary_shape

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name", "Name", ""),
                    ("boundary_shape"      , "", "" ),
            ] )

    def get_name(self):
        """
        returns the optical element name.

        Returns
        -------
        str

        """
        return self._name

    def get_boundary_shape(self):
        """
        Returns the boundary shape.

        Returns
        -------
        None or instance of BoundaryShape

        """
        return self._boundary_shape



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2023, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2023. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #

import numpy

from syned.tools.benders.bender_io import BenderOuputData, BenderFitParameters, BenderStructuralParameters, BenderMovement

class AbstractBenderManager:
    def __init__(self, bender_structural_parameters : BenderStructuralParameters):
        self._bender_structural_parameters = bender_structural_parameters

    @property
    def bender_structural_parameters(self): return self._bender_structural_parameters

    def fit_bender_at_focus_position(self, bender_fit_parameters: BenderFitParameters) -> BenderOuputData: raise NotImplementedError
    def get_bender_shape_from_movement(self, bender_movement: BenderMovement) ->  BenderOuputData: raise NotImplementedError

    def get_q_upstream(self, bender_movement : BenderMovement):   raise NotImplementedError
    def get_q_downstream(self, bender_movement : BenderMovement): raise NotImplementedError

    def calculate_ideal_surface(self, q=None, sign=-1):
        c1, c2, c3, c4, c5, c6, c7, c8, c9, c10 = self.get_conic_coefficients(q)

        x = numpy.linspace(-self.bender_structural_parameters.dim_x_minus, self.bender_structural_parameters.dim_x_plus, self.bender_structural_parameters.bender_bin_x + 1)
        y = numpy.linspace(-self.bender_structural_parameters.dim_y_minus, self.bender_structural_parameters.dim_y_plus, self.bender_structural_parameters.bender_bin_y + 1)

        xx, yy = numpy.meshgrid(x, y)

        c = c1 * (xx ** 2) + c2 * (yy ** 2) + c4 * xx * yy + c7 * xx + c8 * yy + c10
        b = c5 * yy + c6 * xx + c9
        a = c3

        z = (-b + sign * numpy.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
        z[b ** 2 - 4 * a * c < 0] = numpy.nan

        return x, y, z.T

    def get_conic_coefficients(self, q=None):
        theta_grazing = self.bender_structural_parameters.grazing_angle
        ssour         = self.bender_structural_parameters.p
        simag         = q if not q is None else self.bender_structural_parameters.q

        theta = (numpy.pi / 2) - theta_grazing

        ax_maj = (ssour + simag) / 2
        ax_min = numpy.sqrt(simag * ssour) * numpy.cos(theta)
        eccentricity = numpy.sqrt(ax_maj ** 2 - ax_min ** 2) / ax_maj
        #
        # The center is computed on the basis of the object and image positions
        #
        y_center = (ssour - simag) * 0.5 / eccentricity
        z_center = -numpy.sqrt(1 - y_center ** 2 / ax_maj ** 2) * ax_min
        #
        # Computes now the normal in the mirror center.
        #
        rn_center = numpy.zeros(3)
        rn_center[0] = 0.0
        rn_center[1] = -2 * y_center / ax_maj ** 2
        rn_center[2] = -2 * z_center / ax_min ** 2
        rn_center /= numpy.sqrt((rn_center ** 2).sum())
        #
        # Computes the tangent versor in the mirror center.
        #
        rt_center = numpy.zeros(3)
        rt_center[0] = 0.0
        rt_center[1] = rn_center[2]
        rt_center[2] = -rn_center[1]

        # Computes now the quadric coefficient with the mirror center
        # located at (0,0,0) and normal along (0,0,1)

        A = 1 / ax_min ** 2
        B = 1 / ax_maj ** 2
        C = A

        c1 = 0.0  # A if ellipsoid and 0 if cylinder -> managed benders are only cylinders
        c2 = B * rt_center[1] ** 2 + C * rt_center[2] ** 2
        c3 = B * rn_center[1] ** 2 + C * rn_center[2] ** 2
        c4 = 0.0
        c5 = 2 * (B * rn_center[1] * rt_center[1] + C * rn_center[2] * rt_center[2])
        c6 = 0.0
        c7 = 0.0
        c8 = 0.0
        c9 = 2 * (B * y_center * rn_center[1] + C * z_center * rn_center[2])
        c10 = 0.0

        return c1, c2, c3, c4, c5, c6, c7, c8, c9, c10

class CalibrationParameters:
    __parameters_upstream   = [0.0, 0.0]
    __parameters_downstream = [0.0, 0.0]

    def __init__(self, parameters_upstream, parameters_downstream):
        self.__parameters_upstream   = parameters_upstream
        self.__parameters_downstream = parameters_downstream

    @property
    def upstream(self):   return self.__parameters_upstream
    @property
    def downstream(self): return self.__parameters_downstream

class StandardBenderManager(AbstractBenderManager):
    def __init__(self, bender_structural_parameters : BenderStructuralParameters):
        super(StandardBenderManager, self).__init__(bender_structural_parameters=bender_structural_parameters)

    def get_q_upstream(self, bender_movement : BenderMovement):   return bender_movement.position_upstream # position coincides with q -> direct calculation
    def get_q_downstream(self, bender_movement : BenderMovement): return bender_movement.position_downstream

class CalibratedBenderManager(AbstractBenderManager):
    def __init__(self, bender_structural_parameters : BenderStructuralParameters, calibration_parameters : CalibrationParameters):
        super(CalibrationParameters, self).__init__(bender_structural_parameters=bender_structural_parameters)
        self._calibration_parameters       = calibration_parameters


    def get_q_upstream(self, bender_movement : BenderMovement):
        if bender_movement.position_upstream is None: return None
        return CalibratedBenderManager.__get_q_from_calibration(bender_movement.position_upstream,
                                                                self._calibration_parameters.upstream)

    def get_q_downstream(self, bender_movement : BenderMovement):
        if bender_movement.position_downstream is None: return None
        return CalibratedBenderManager.__get_q_from_calibration(bender_movement.position_downstream,
                                                                self._calibration_parameters.downstream)

    @classmethod # 1/q = p0*pos + p1
    def __get_q_from_calibration(cls, position, parameters):  return 1 / (parameters[0] * position + parameters[1])


import copy
from collections import OrderedDict
try:
    import json_tricks as json # to save numpy arrays
except:
    import json

# TODO: although basic functionality is implemented, the use of exec should be replace by introspection tools
class SynedObject(object):
    """
    This is the base object for SYNED.

    It includes the methods of the common interface to allow json file input/output and info mechanism

    These standard methods are:
        * keys()
        * to_dictionary()
        * to_full_dictionary()
        * to_json()
        * info()
        * set_value_from_key_name()
        * get_value_from_key_name()

    """

    def _set_support_text(self, text):
        ordered_support_dict = OrderedDict()
        for e in text:
            ordered_support_dict[e[0]] = (e[1], e[2])
        self._support_dictionary = ordered_support_dict

    def _add_support_text(self, text):
        try:
            for e in text:
                self._support_dictionary[e[0]] = (e[1], e[2])
        except:
            self._set_support_text(text)

    #
    # this is the common interface to allow json file input/output and info mechanism
    #
    # standard methods are:
    #   keys
    #   to_dictionary
    #   to_full_dictionary
    #   to_json
    #   info
    #   set_value_from_key_name
    #   get_value_from_key_name
    #
    def keys(self):
        """
        Returns the keys of the supporting doctionary.
        Returns
        -------
        list
            A list of keys.

        """
        try:
            return self._support_dictionary.keys()
        except:
            return None

    def to_dictionary(self):
        """
        Returns a dictionary with the object fields.

        Some dictionary keys may contain SYNED instances. Use to_full_dictionary to recurrently expand these objects
        in their basic ingredients.

        Returns
        -------
        dict
            A dictionary with the data.

        """
        dict_to_save = OrderedDict()
        dict_to_save.update({"CLASS_NAME":self.__class__.__name__})

        try:
            if self.keys() is not None:
                for key in self.keys():
                    tmp1 = eval("self._%s" % (key) )
                    if isinstance(tmp1,SynedObject):
                        dict_to_save[key] = tmp1.to_dictionary()
                    else:
                        dict_to_save[key] = tmp1
        except:
            pass

        return dict_to_save

    def to_full_dictionary(self):
        """
        Returns a dictionary with the object fields, including other syned objects embedded or list of elements.

        The "full" means that the SYNED instances found are recurrently expanded in their basic ingredients.

        Returns
        -------
        dict

        """
        dict_to_save = OrderedDict()
        dict_to_save.update({"CLASS_NAME":self.__class__.__name__})
        try:
            if self.keys() is not None:
                for key in self.keys():
                    tmp1 = eval("self._%s" % (key) )
                    if isinstance(tmp1, SynedObject):
                        dict_to_save[key] = tmp1.to_full_dictionary()
                    else:
                        mylist = []
                        mylist.append(tmp1)
                        mylist.append(self._support_dictionary[key])
                        dict_to_save[key] = mylist # [tmp1,self._support_dictionary[key]]
        except:
            print("** Warning: failed to load/write one or multiple items in ", self.keys())

        return dict_to_save

    def to_json(self, file_name=None):
        """
        Writes a json file with the SYNED object data.

        Parameters
        ----------
        file_name : str
            The file name

        Returns
        -------
        str
            JSON formatted str. The result of json.dumps()

        """
        dict1 = OrderedDict()
        dict1.update(self.to_dictionary())

        jsn1 = json.dumps(dict1, indent=4, separators=(',', ': '))
        if file_name != None:
            f = open(file_name,'w')
            f.write(jsn1)
            f.close()
            print("File written to disk: %s"%(file_name))
        return jsn1

    def info_recurrent(self, fd, prefix="    "):
        """
        Get text info of recurrent SYNED objects.

        Parameters
        ----------
        fd : dict
            The dictionary with SYNED data.
        prefix : str, optional
            Prefix to indent recursive items.

        Returns
        -------
        str

        """
        text = ""
        prefix1 = prefix
        for key in fd.keys():
            if isinstance(fd[key],OrderedDict):
                text += prefix1 + self.info_recurrent(fd[key], prefix=prefix1)
            elif isinstance(fd[key],str):
                text += prefix1 + "-------%s---------\n"%fd[key]
            elif isinstance(fd[key],list):
                if isinstance(fd[key][0],OrderedDict):
                    for element in fd[key]:
                        text += self.info_recurrent(element, prefix=prefix1)
                elif isinstance(fd[key][0],list):
                    for i,element in enumerate(fd[key][0]):
                        try:
                            # text += "****\n"
                            text += prefix1 + element.info()
                        except:
                            text += ("%s%s[%d]: %s\n" %(prefix1, key, i, repr(element))) # used for conic coefficients
                else:
                    text += prefix1 + prefix1 + '%s: %s %s # %s\n' %(key,  repr(fd[key][0]), fd[key][1][1], fd[key][1][0])
            else:
                pass
        return text

    def info(self):
        """
        Get text info of recurrent SYNED objects.

        Returns
        -------
        str

        """
        return self.info_recurrent( self.to_full_dictionary() )

    def set_value_from_key_name(self,key,value):
        """
        Sets a value using its key value.

        Parameters
        ----------
        key : str
            The key for the value to modify.
        value
            The new value

        """
        if key in self.keys():
            try:
                exec("self._%s = value" % (key))
                # print("Set variable %s to value: "%key + repr(value))
            except:
                raise ValueError("Cannot set variable %s to value: "%key + repr(value) )
        else:
            raise ValueError("Key %s not accepted by class %s"%(key,self.__class__.__name__))

    def get_value_from_key_name(self, key):
        """
        Gets a value using its key value.

        Parameters
        ----------
        key : str
            The key for the value to retrieve.

        """
        try:
            value = eval("self._%s" % (key))
            return value
        except:
            raise ValueError("Cannot get variable %s: "%key)

    def duplicate(self):
        """
        Returns a copy of the SYNED object instance.

        Returns
        -------
        SynedObject instance
            A copy of the object instance.

        """
        return copy.deepcopy(self)



"""
Implement an undulator with vertical and horizontal magnetic fields.
"""
import numpy
import scipy.constants as codata

cte = codata.e/(2*numpy.pi*codata.electron_mass*codata.c)

from syned.storage_ring.magnetic_structures.insertion_device import InsertionDevice

class Undulator(InsertionDevice):
    """
    Constructor.

    Parameters
    ----------
    K_vertical : float, optional
        The deflection K parameter corresponding to magnetic field in the vertical direction.
    K_horizontal : float, optional
        The deflection K parameter corresponding to magnetic field in the horizontal direction.
    period_length : float, optional
        The ID period in m.
    number_of_periods : float, optional
        The number of periods. It may be a float, considering that number_of_periods = ID_length / period_length.

    """

    def __init__(self,
                 K_vertical = 0.0,
                 K_horizontal = 0.0,
                 period_length = 0.0,
                 number_of_periods = 1.0):
        InsertionDevice.__init__(self, K_vertical, K_horizontal, period_length, number_of_periods)

    def resonance_wavelength(self, gamma, theta_x=0.0, theta_z=0.0, harmonic=1.0):
        """
        Returns the resonant wavelength in m.

        Parameters
        ----------
        gamma : float
            The gamma or Lorentz factor of the electron beam.
        theta_x : float, optional
            The angle along the horizontal direction in rad.
        theta_z : float, optional
            The angle along the vertical direction in rad.
        harmonic : float, optional
            The harmonic number (1,2,3...) to be considered.

        Returns
        -------
        float


        """
        wavelength = (self.period_length() / (2.0 * gamma **2)) * \
                     (1 + self.K_vertical()**2 / 2.0 + self.K_horizontal()**2 / 2.0 + \
                      gamma**2 * (theta_x**2 + theta_z**2))

        return wavelength / harmonic

    def resonance_frequency(self, gamma, theta_x=0.0, theta_z=0.0, harmonic=1.0):
        """
        Returns the resonant frequency in Hz.

        Parameters
        ----------
        gamma : float
            The gamma or Lorentz factor of the electron beam.
        theta_x : float, optional
            The angle along the horizontal direction in rad.
        theta_z : float, optional
            The angle along the vertical direction in rad.
        harmonic : float, optional
            The harmonic number (1,2,3...) to be considered.

        Returns
        -------
        float


        """
        frequency = codata.c / self.resonance_wavelength(gamma, theta_x, theta_z)

        return frequency * harmonic

    def resonance_energy(self, gamma, theta_x=0.0, theta_z=0.0, harmonic=1.0):
        """
        Returns the resonant energy in eV.

        Parameters
        ----------
        gamma : float
            The gamma or Lorentz factor of the electron beam.
        theta_x : float, optional
            The angle along the horizontal direction in rad.
        theta_z : float, optional
            The angle along the vertical direction in rad.
        harmonic : float, optional
            The harmonic number (1,2,3...) to be considered.

        Returns
        -------
        float


        """
        energy_in_ev = codata.h * self.resonance_frequency(gamma, theta_x, theta_z) / codata.e

        return energy_in_ev*harmonic

    def gaussian_central_cone_aperture(self, gamma, n=1.0):
        """
        Returns the approximated angular width (sigma) in rad of the central cone.

        Parameters
        ----------
        gamma : float
            The gamma or Lorentz factor of the electron beam.
        n : float, optional
            The harmonic number (1,2,3...) to be considered.

        Returns
        -------
        float

        References
        ----------
        Eq. 15 in https://xdb.lbl.gov/Section2/Sec_2-1.html

        """
        return (1/gamma)*numpy.sqrt((1.0/(2.0*n*self.number_of_periods())) * (1.0 + self.K_horizontal()**2/2.0 + self.K_vertical()**2/2.0))

    @classmethod
    def initialize_as_vertical_undulator(cls, K = 0.0, period_length = 0.0, periods_number = 1.0, **params):
        """
        Create an undulator with vertical magnetic field.

        Parameters
        ----------
        K : float, optional
            The deflection K parameter corresponding to magnetic field in the vertical direction.
        period_length : float, optional
            The ID period in m.
        periods_number : float, optional
            The number of periods. It may be a float, considering that number_of_periods = ID_length / period_length.
        params : other parameters accepted by Undulator.

        Returns
        -------
        instance of Undulator

        """
        return cls(K_vertical=K,
                   K_horizontal=0.0,
                   period_length=period_length,
                   number_of_periods=periods_number, **params)
    #
    #
    #

    def get_sigmas_radiation(self, gamma, harmonic=1.0):
        """
        Returns the approximated radiation sigmas (size and divergence).

        Parameters
        ----------
        gamma : float
            The gamma or Lorentz factor of the electron beam.
        harmonic : float, optional
            The harmonic number (1,2,3...) to be considered.

        Returns
        -------
        tuple
            (sigma_in_real_space [m], sigma_in_divergence_space [rad])

        References
        ----------

        See formulas 25 & 30 in Elleaume (Onuki & Elleaume) for the calculated sizes of the photon undulator beam.
        Onuki & Elleaume "Undulators, Wigglers and Their Applications" (2002) CRC Press.


        """
        # calculate sizes of the photon undulator beam
        # see formulas 25 & 30 in Elleaume (Onuki & Elleaume)
        photon_energy = self.resonance_energy(gamma,harmonic=harmonic)
        lambdan = 1e-10 * codata.h * codata.c / codata.e * 1e10 / photon_energy  # in m
        sigma_r = 2.740 / 4 / numpy.pi * numpy.sqrt(lambdan * self.number_of_periods() * self.period_length())
        sigma_r_prime =  0.69 * numpy.sqrt(lambdan / (self.length()))
        return sigma_r, sigma_r_prime

    def get_resonance_ring(self, gamma, harmonic=1.0, ring_order=1):
        """
        Return the angular position of the rings at resonance.

        Parameters
        ----------
        gamma : float
            The gamma or Lorentz factor of the electron beam.
        harmonic : float, optional
            The harmonic number (1,2,3...) to be considered.
        ring_order : int, optional
            The ring order (1,2,3...)

        Returns
        -------
        float
            The angular position in rad.

        """
        K_value = numpy.sqrt( self.K_vertical()**2 + self.K_horizontal()**2)
        return 1.0 / gamma * numpy.sqrt( ring_order / harmonic * (1 + 0.5 * K_value**2) )


    def undulator_full_emitted_power(self, gamma, ring_current):
        """
        Returns the total emitted power by an undulator.

        Parameters
        ----------
        gamma : float
            The gamma or Lorentz factor of the electron beam.
        ring_current : float
            The current of the electron beam in A.

        Returns
        -------
        float
            The power in W.

        References
        ----------
        Eq. 18 in https://xdb.lbl.gov/Section2/Sec_2-1.html

        """
        ptot = (self.number_of_periods() / 6) * codata.value('characteristic impedance of vacuum') * \
               ring_current * codata.e * 2 * numpy.pi * codata.c * gamma**2 * \
               (self.K_vertical()**2 + self.K_horizontal()**2) / self.period_length()
        return ptot

    def get_photon_sizes_and_divergences(self, syned_electron_beam, harmonic=1):
        """
        Return the photon beam sizes and divergences.

        They are calculates as a convolution of the radiation sigmas (get_sigmas_radiation()) and
        the electron beam sizes.

        Parameters
        ----------
        syned_electron_beam : instance of ElectronBeam
            The electron bean used to get the electron beam sizes and electron energy.
        harmonic : int
            The harmonic number (1,2,3...).

        Returns
        -------
        tuple
            (Sx [m], Sz [m], Sxp [rad], Szp [rad])

        """
        sr,srp = self.get_sigmas_radiation(syned_electron_beam.gamma(), harmonic=harmonic)
        sx,sxp,sz,szp = syned_electron_beam.get_sigmas_all()

        Sx = numpy.sqrt( sx**2 + sr**2)
        Sz = numpy.sqrt( sz**2 + sr**2)
        Sxp = numpy.sqrt( sxp**2 + srp**2)
        Szp = numpy.sqrt( szp**2 + srp**2)

        return Sx,Sz,Sxp,Szp

    def get_K_from_photon_energy(self, photon_energy, gamma, harmonic=1):
        """
        Calculate K for a given resonance energy.

        Parameters
        ----------
        photon_energy : float
            The photon energy in eV.
        gamma : float
            The electron gamma (Lorentz factor).
        harmonic : int
            The harmonic number.

        Returns
        -------
        float

        """
        m2ev = codata.c * codata.h / codata.e
        wavelength = harmonic * m2ev / photon_energy
        return numpy.sqrt(2 * (((wavelength * 2 * gamma**2) / self.period_length()) - 1))

    def approximated_coherent_fraction_horizontal(self, syned_electron_beam, harmonic=1):
        """
        Gets the approximated coherent fraction at resonance in the horizontal direction.

        It is calculated as the ratio of the phase space of the coherent emission (zero emittance) over
        the actual emission (finite emittance).

        Parameters
        ----------
        syned_electron_beam : instance of ElectronBeam
            The electron bean used to get the electron beam sizes and electron energy.
        harmonic : int
            The harmonic number (1,2,3...).

        Returns
        -------
        float

        """
        Sx,Sy,Sxp,Syp = self.get_photon_sizes_and_divergences(syned_electron_beam, harmonic=harmonic)
        srad,sradp = self.get_sigmas_radiation(syned_electron_beam.gamma(), harmonic=harmonic)
        return srad * sradp / ( Sx * Sxp)

    def approximated_coherent_fraction_vertical(self, syned_electron_beam,harmonic=1):
        """
        Gets the approximated coherent fraction at resonance in the vertical direction.

        It is calculated as the ratio of the phase space of the coherent emission (zero emittance) over
        the actual emission (finite emittance).

        Parameters
        ----------
        syned_electron_beam : instance of ElectronBeam
            The electron bean used to get the electron beam sizes and electron energy.
        harmonic : int
            The harmonic number (1,2,3...).

        Returns
        -------
        float

        """
        Sx,Sy,Sxp,Syp = self.get_photon_sizes_and_divergences(syned_electron_beam, harmonic=harmonic)
        srad,sradp = self.get_sigmas_radiation(syned_electron_beam.gamma(), harmonic=harmonic)
        return srad * sradp / ( Sy * Syp)

    def approximated_coherent_fraction(self, syned_electron_beam, harmonic=1):
        """
        Gets the approximated coherent fraction at resonance.

        It is calculated as the product of the horizontal and vertical coherent fractions.

        Parameters
        ----------
        syned_electron_beam : instance of ElectronBeam
            The electron bean used to get the electron beam sizes and electron energy.
        harmonic : int
            The harmonic number (1,2,3...).

        Returns
        -------
        float

        """
        return self.approximated_coherent_fraction_horizontal(syned_electron_beam, harmonic=harmonic) * \
               self.approximated_coherent_fraction_vertical(syned_electron_beam, harmonic=harmonic)


if __name__ == "__main__":

    a = Undulator(number_of_periods=61.5, period_length=0.057)
    a.set_K_vertical_from_magnetic_field(0.187782)

    print(a._K_vertical)

    print (a.resonance_energy(gamma=5870.8540997356595))

    fd = a.to_full_dictionary()
    dict = a.to_dictionary()

    print(dict)

    for key in fd:
        print(key,fd[key][0])

    for key in fd:
        print(key,dict[key])

    print(a.keys())
    print(a.info())


    print("####### derived quantities ##########")

    from syned.storage_ring.electron_beam import ElectronBeam
    ebeam = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=2.0,energy_spread=0.0,current=0.5)

    sigmas_radiation = a.get_sigmas_radiation(ebeam.gamma(),harmonic=1.0)
    print("sigmas radiation [m rad]:",sigmas_radiation)


    ring = a.get_resonance_ring(ebeam.gamma(), harmonic=1.0, ring_order=1)
    print("first ring at [rad]:",ring)

    pow = a.undulator_full_emitted_power(ebeam.gamma(), ebeam.current())
    print("Total emission [W]",pow)

    sizes = a.get_photon_sizes_and_divergences(ebeam,harmonic=1)
    print("Sizes: ",sizes)

    print("Resonance: ",a.resonance_energy(ebeam.gamma()))

    print("K at 444 eV",a.get_K_from_photon_energy(444.0,ebeam.gamma()))
    print("K at 200 eV", a.get_K_from_photon_energy(200.0, ebeam.gamma()))

    print("CF H V HV: ",
          a.approximated_coherent_fraction_horizontal(ebeam,harmonic=1),
          a.approximated_coherent_fraction_vertical(ebeam, harmonic=1),
          a.approximated_coherent_fraction(ebeam, harmonic=1),)



#
# this is a filter with a hole (holed-filter)
#                                  Slit     BeamStopper    Filter      HoledFilter
#       beam pass at center         Yes     No             Yes         No
#       apply attenuation           No      No             Yes         Yes
#
#
from syned.beamline.optical_elements.absorbers.absorber import Absorber

class HoledFilter(Absorber):
    """
    Filter or absorber or attenuator with a hole.

    Note that:
                                   Slit     BeamStopper    Filter      HoledFilter
        beam pass at center         Yes     No             Yes         No
        apply attenuation           No      No             Yes         Yes

    Constructor.

    Parameters
    ----------
    name : str
        The name of the optical element.
    material : str
        A string defining the material.
    thickness : float
        The filter thickness in m.
    boundary_shape : instance of BoundaryShape, optional
        Defines the geometry of the hole. if None, it is initialized to BoundaryShape().
    """
    def __init__(self,
                 name="Undefined",
                 material="Be",
                 thickness=1e-3,
                 boundary_shape=None):
        Absorber.__init__(self, name=name, boundary_shape=boundary_shape)
        self._material = material
        self._thickness = thickness


        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("material"      , "Material (symbol, formula or name)",    "" ),
                    ("thickness"     , "Thickness ",                            "m" ),
            ] )
    def get_material(self):
        """
        Returns the material name.

        Returns
        -------
        str

        """
        return self._material

    def get_thickness(self):
        """
        Retuirns the filter thickness in m.

        Returns
        -------
        float

        """
        return self._thickness



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2023, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2023. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #

import numpy
from scipy.interpolate import interp2d
from scipy.optimize import curve_fit
from scipy import integrate

from syned.tools.benders.bender_io import BenderMovement, BenderFitParameters, BenderStructuralParameters, BenderOuputData
from syned.tools.benders.bender_manager import StandardBenderManager, CalibratedBenderManager, CalibrationParameters

class FixedRodsBenderFitParameters(BenderFitParameters):
    def __init__(self,
                 optimized_length=None,
                 n_fit_steps=None,
                 R0=None,
                 R0_max=False,
                 R0_min=None,
                 R0_fixed=None,
                 eta=None,
                 eta_max=False,
                 eta_min=None,
                 eta_fixed=None,
                 W2=None,
                 W2_max=False,
                 W2_min=None,
                 W2_fixed=None):
        super(FixedRodsBenderFitParameters, self).__init__(optimized_length=optimized_length,                                                     
                                                               n_fit_steps=n_fit_steps)
        self.R0 = R0
        self.R0_max = R0_max
        self.R0_min = R0_min
        self.R0_fixed = R0_fixed
        self.eta = eta
        self.eta_max = eta_max
        self.eta_min = eta_min
        self.eta_fixed = eta_fixed
        self.W2 = W2
        self.W2_max = W2_max
        self.W2_min = W2_min
        self.W2_fixed = W2_fixed

class FixedRodsBenderOuputData(BenderOuputData):
    def __init__(self,
                 x=None,
                 y=None,
                 ideal_profile=None,
                 bender_profile=None,
                 correction_profile=None,
                 titles=None,
                 z_bender_correction=None,
                 z_figure_error=None,
                 z_bender_correction_no_figure_error=None,
                 R0_out=None,
                 eta_out=None,
                 W2_out=None,
                 alpha= None,
                 W0=None,
                 F_upstream=None,
                 F_downstream=None):
        super(FixedRodsBenderOuputData, self).__init__(x,
                                                 y,
                                                 ideal_profile,
                                                 bender_profile,
                                                 correction_profile,
                                                 titles,
                                                 z_bender_correction,
                                                 z_figure_error,
                                                 z_bender_correction_no_figure_error)
        self.R0_out       = R0_out
        self.eta_out      = eta_out
        self.W2_out       = W2_out
        self.alpha        = alpha
        self.W0           = W0
        self.F_upstream   = F_upstream
        self.F_downstream = F_downstream


class FixedRodsBenderStructuralParameters(BenderStructuralParameters):
    def __init__(self,
                 dim_x_minus=None,
                 dim_x_plus=None,
                 bender_bin_x=None,
                 dim_y_minus=None,
                 dim_y_plus=None,
                 bender_bin_y=None,
                 p=None,
                 q=None,
                 grazing_angle=None,
                 E=None,
                 h=None,
                 figure_error_mesh=None,
                 r=None,
                 l=None,
                 R0=None,
                 eta=None,
                 W2=None,
                 alpha=None,
                 W0=None,
                 workspace_units_to_m=None,
                 workspace_units_to_mm=None):
        super(FixedRodsBenderStructuralParameters, self).__init__(dim_x_minus=dim_x_minus,
                                                                      dim_x_plus=dim_x_plus ,
                                                                      bender_bin_x=bender_bin_x ,
                                                                      dim_y_minus=dim_y_minus,
                                                                      dim_y_plus=dim_y_plus,
                                                                      bender_bin_y=bender_bin_y,
                                                                      p=p,
                                                                      q=q,
                                                                      grazing_angle=grazing_angle,
                                                                      E=E,
                                                                      h=h,
                                                                      figure_error_mesh=figure_error_mesh,
                                                                      workspace_units_to_m=workspace_units_to_m,
                                                                      workspace_units_to_mm=workspace_units_to_mm)
        self.r     = r
        self.l     = l
        self.R0    = R0
        self.eta   = eta
        self.W2    = W2
        self.alpha = alpha
        self.W0    = W0

epsilon_minus = 1 - 1e-8
epsilon_plus  = 1 + 1e-8

# Decorator
class _FixedRodsBenderCalculator():
    def __init__(self, bender_manager):
        self.__bender_manager = bender_manager

    def fit_bender_at_focus_position(self, bender_fit_parameters: FixedRodsBenderFitParameters) -> FixedRodsBenderOuputData:
        workspace_units_to_m  = self.__bender_manager.bender_structural_parameters.workspace_units_to_m
        workspace_units_to_mm = self.__bender_manager.bender_structural_parameters.workspace_units_to_mm
        optimized_length      = bender_fit_parameters.optimized_length

        x             = numpy.linspace(-self.__bender_manager.bender_structural_parameters.dim_x_minus, self.__bender_manager.bender_structural_parameters.dim_x_plus, self.__bender_manager.bender_structural_parameters.bender_bin_x + 1)
        y             = numpy.linspace(-self.__bender_manager.bender_structural_parameters.dim_y_minus, self.__bender_manager.bender_structural_parameters.dim_y_plus, self.__bender_manager.bender_structural_parameters.bender_bin_y + 1)
        W1            = self.__bender_manager.bender_structural_parameters.dim_x_plus + self.__bender_manager.bender_structural_parameters.dim_x_minus
        L             = self.__bender_manager.bender_structural_parameters.dim_y_plus + self.__bender_manager.bender_structural_parameters.dim_y_minus
        p             = self.__bender_manager.bender_structural_parameters.p
        q             = self.__bender_manager.bender_structural_parameters.q
        grazing_angle = self.__bender_manager.bender_structural_parameters.grazing_angle
    
        E = self.__bender_manager.bender_structural_parameters.E
        l = self.__bender_manager.bender_structural_parameters.l
        h = self.__bender_manager.bender_structural_parameters.h
        r = self.__bender_manager.bender_structural_parameters.r

        if optimized_length is None:
            y_fit = y
        else:
            cursor = numpy.where(numpy.logical_and(y >= -optimized_length / 2, y <= optimized_length / 2))
            y_fit = y[cursor]
    
        ideal_slope_profile_fit = ideal_slope_profile(y_fit, p, q, grazing_angle)

        initial_guess    = [bender_fit_parameters.R0, bender_fit_parameters.eta, bender_fit_parameters.W2]
        constraints     =  [[bender_fit_parameters.R0_min if bender_fit_parameters.R0_fixed == False else (bender_fit_parameters.R0 * epsilon_minus),
                             bender_fit_parameters.eta_min if bender_fit_parameters.eta_fixed == False else (bender_fit_parameters.eta * epsilon_minus),
                             bender_fit_parameters.W2_min if bender_fit_parameters.W2_fixed == False else (bender_fit_parameters.W2 * epsilon_minus)],
                            [bender_fit_parameters.R0_max if bender_fit_parameters.R0_fixed == False else (bender_fit_parameters.R0 * epsilon_plus),
                             bender_fit_parameters.eta_max if bender_fit_parameters.eta_fixed == False else (bender_fit_parameters.eta * epsilon_plus),
                             bender_fit_parameters.W2_max if bender_fit_parameters.W2_fixed == False else (bender_fit_parameters.W2 * epsilon_plus)]
                           ]
    
        def bender_function(x, R0, eta, W2):
            return bender_slope_profile(x, p, q, grazing_angle, W1, L, R0 / workspace_units_to_m, eta, W2 / workspace_units_to_mm)
    
        for i in range(bender_fit_parameters.n_fit_steps):
            parameters, _ = curve_fit(f=bender_function,
                                      xdata=y_fit,
                                      ydata=ideal_slope_profile_fit,
                                      p0=initial_guess,
                                      bounds=constraints,
                                      method='trf')
            initial_guess = parameters
    
        R0  = parameters[0] / workspace_units_to_m # here in workspace units
        eta = parameters[1]
        W2  = parameters[2] / workspace_units_to_mm
    
        alpha = calculate_taper_factor(W1, W2, L, p, q, grazing_angle)
        W0    = calculate_W0(W1, alpha, L, p, q, grazing_angle) # W at the center
    
        bender_profile = bender_height_profile(y, p, q, grazing_angle, R0, eta, alpha)
    
        F_upstream, F_downstream = calculate_bender_forces(q, R0, eta, E, W0, l, h, r)
    
        parameters = numpy.append(parameters, round(alpha, 3))
        parameters = numpy.append(parameters, round(W0 * workspace_units_to_mm, 4))
        parameters = numpy.append(parameters, round(F_upstream, 6))
        parameters = numpy.append(parameters, round(F_downstream, 6))
    
        ideal_profile  = ideal_height_profile(y, p, q, grazing_angle)
    
        # back to Shadow system
        bender_profile -= numpy.min(bender_profile)
        ideal_profile  -= numpy.min(ideal_profile)
    
        # from here it's Shadow Axis system
        correction_profile = ideal_profile - bender_profile
        if not optimized_length is None: correction_profile_fit = correction_profile[cursor]
    
        # r-squared = 1 - residual sum of squares / total sum of squares
        r_squared = 1 - (numpy.sum(correction_profile ** 2) / numpy.sum((ideal_profile - numpy.mean(ideal_profile)) ** 2))
        rms       = round(correction_profile.std() * 1e9 * workspace_units_to_m, 6)
        if not bender_fit_parameters.optimized_length is None: rms_opt = round(correction_profile_fit.std() * 1e9 * workspace_units_to_m, 6)
    
        z_bender_correction = numpy.zeros((len(x), len(y)))
    
        for i in range(z_bender_correction.shape[0]): z_bender_correction[i, :] = numpy.copy(correction_profile)
    
        bender_data = FixedRodsBenderOuputData(x=x,
                                               y=y,
                                               ideal_profile=ideal_profile,
                                               bender_profile=bender_profile,
                                               correction_profile=correction_profile,
                                               titles=["Bender vs. Ideal Profiles" + "\n" + r'$R^2$ = ' + str(r_squared),
                                                  "Correction Profile 1D, r.m.s. = " + str(rms) + " nm" +
                                                  ("" if optimized_length is None else (", " + str(rms_opt) + " nm (optimized)"))],
                                               z_bender_correction_no_figure_error=z_bender_correction)
    
        if not self.__bender_manager.bender_structural_parameters.figure_error_mesh is None:
            x_e, y_e, z_e = self.__bender_manager.bender_structural_parameters.figure_error_mesh
    
            if len(x) == len(x_e) and len(y) == len(y_e) and \
                    x[0] == x_e[0] and x[-1] == x_e[-1] and \
                    y[0] == y_e[0] and y[-1] == y_e[-1]:
                z_figure_error = z_e
            else:
                z_figure_error = interp2d(y_e, x_e, z_e, kind='cubic')(y, x)
    
            bender_data.z_figure_error      = z_figure_error
            bender_data.z_bender_correction = bender_data.z_bender_correction_no_figure_error + z_figure_error
        else:
            bender_data.z_bender_correction = bender_data.z_bender_correction_no_figure_error
    
        bender_data.R0_out       = round(parameters[0], 5)
        bender_data.eta_out      = parameters[1]
        bender_data.W2_out       = round(parameters[2], 3)
        bender_data.alpha        = parameters[3]
        bender_data.W0           = parameters[4]
        bender_data.F_upstream   = parameters[5]
        bender_data.F_downstream = parameters[6]
    
        # set the structure of the mirror at focus
        self.__bender_manager.bender_structural_parameters.R0    = bender_data.R0_out
        self.__bender_manager.bender_structural_parameters.eta   = bender_data.eta_out
        self.__bender_manager.bender_structural_parameters.W2    = bender_data.W2_out
        self.__bender_manager.bender_structural_parameters.alpha = bender_data.alpha
        self.__bender_manager.bender_structural_parameters.W0    = bender_data.W0
        
        return bender_data

    def get_bender_shape_from_movement(self, bender_movement: BenderMovement) ->  FixedRodsBenderOuputData:
        pass


class FixedRodsStandardBenderManager(StandardBenderManager):
    def __init__(self,
                 bender_structural_parameters : FixedRodsBenderStructuralParameters):
        super(FixedRodsStandardBenderManager, self).__init__(bender_structural_parameters=bender_structural_parameters)
        self.__calculator = _FixedRodsBenderCalculator(bender_manager=self)

    def fit_bender_at_focus_position(self, bender_fit_parameters: FixedRodsBenderFitParameters) -> FixedRodsBenderOuputData:
        return self.__calculator.fit_bender_at_focus_position(bender_fit_parameters)

    def get_bender_shape_from_movement(self, bender_movement: BenderMovement) ->  FixedRodsBenderOuputData:
        return self.__calculator.get_bender_shape_from_movement(bender_movement)


class FixedRodsCalibratedBenderManager(CalibratedBenderManager):
    def __init__(self,
                 bender_structural_parameters: FixedRodsBenderStructuralParameters,
                 calibration_parameters : CalibrationParameters):
        super(FixedRodsCalibratedBenderManager, self).__init__(bender_structural_parameters=bender_structural_parameters,
                                                         calibration_parameters=calibration_parameters)
        self.__calculator = _FixedRodsBenderCalculator(bender_manager=self)

    def fit_bender_at_focus_position(self, bender_fit_parameters: FixedRodsBenderFitParameters) -> FixedRodsBenderOuputData:
        return self.__calculator.fit_bender_at_focus_position(bender_fit_parameters)

    def get_bender_shape_from_movement(self, bender_movement: BenderMovement) -> FixedRodsBenderOuputData:
        return self.__calculator.get_bender_shape_from_movement(bender_movement)




def focal_distance(p, q):
    return p*q/(p+q)

def demagnification_factor(p, q):
    return p/q

def mu_nu(m):
    return (m - 1) / (m + 1), m/(m+1)**2

def calculate_ideal_slope_variation(y, fprime_, K0id, mu, nu):
    return 2*fprime_*K0id*((2 * nu * (y / fprime_) + mu) / numpy.sqrt(1 - mu * (y / fprime_) - nu * (y / fprime_) ** 2) - mu)

def fprime(p, q, grazing_angle):
    return focal_distance(p, q) / numpy.cos(grazing_angle)

def calculate_bender_slope_variation(y, fprime_, K0, eta, alpha):
    return -(K0*fprime_/alpha**2)*(eta * alpha * (y / fprime_) + (eta + alpha) * numpy.log(1 - (alpha * y / fprime_)))

def calculate_taper_factor(W1, W2, L, p, q, grazing_angle):
    # W2 = W1(1 - alpha L/f')
    # W2/W1 - 1 = - alpha  L/f'
    # f'/L ( 1 - W2/W1) = alpha
    return (1 - W2/W1) * (fprime(p, q, grazing_angle) / L)

def calculate_W0(W1, alpha, L, p, q, grazing_angle):
    return W1*(1 - alpha * L / (2 * fprime(p, q, grazing_angle)))

def ideal_slope_profile(y, p, q, grazing_angle):
    mu, nu  = mu_nu(demagnification_factor(p, q))
    fprime_ = fprime(p, q, grazing_angle)
    K0id    = numpy.tan(grazing_angle)/(2*fprime_)

    return calculate_ideal_slope_variation(y, fprime_, K0id, mu, nu)

def ideal_height_profile(y, p, q, grazing_angle):
    mu, nu  = mu_nu(demagnification_factor(p, q))
    fprime_ = fprime(p, q, grazing_angle)
    K0id    = numpy.tan(grazing_angle)/(2*fprime_)

    profile = numpy.zeros(len(y))
    for i in range(len(y)): profile[i] = integrate.quad(func=(lambda x: calculate_ideal_slope_variation(x, fprime_, K0id, mu, nu)), 
                                                        a=y[0], b=y[i])[0]

    return profile

def bender_slope_profile(y, p, q, grazing_angle, W1, L, R0, eta, W2):
    return calculate_bender_slope_variation(y, fprime(p, q, grazing_angle), 1 / R0, eta, alpha=calculate_taper_factor(W1, W2, L, p, q, grazing_angle))

def bender_height_profile(y, p, q, grazing_angle, R0, eta, alpha):
    fprime_ = fprime(p, q, grazing_angle)

    profile = numpy.zeros(len(y))
    for i in range(len(y)): profile[i] = integrate.quad(func=(lambda x: calculate_bender_slope_variation(x, fprime_, 1/R0, eta, alpha)), 
                                                        a=y[0], b=y[i])[0]

    return profile

# -----------------------------------------------
# q = focus distance (from mirror center) (1/p + 1/q = 1/f lenses equation)
# eta = bender asymmetry factor (from slope minimization)
# K0 = 1/R (radius of curvature at the center)
# E0 = Young's modulus
# L = length of the mirror
# r = distance between inner ond outer rods
# ----------------------------------------------------------------
def calculate_bender_forces(q, R0, eta, E, W0, L, h, r):
    I0 = (W0*h**3)/12
    M0 = E*I0/R0
    F_upstream = (M0/r) * (1 - (eta*(L + 2*r)/(2*q)))
    F_downstream = (M0/r) * (1 + (eta*(L + 2*r)/(2*q)))

    return F_upstream, F_downstream
